#include "MTCP.h"

void
MTCP::Init() {
  IPC::Init(MTCP_SERVER::Init);
}

pid_t
MTCP::Fork() {
  pid_t pid = fork();
  if (pid == 0) { /* Child. */
    /* Obtain IPC connection. */
    IPC::ObtainConnection(CONNECTION_IPC);
    /* Annotate duplicated socket connections with pid. */
    IPC::Reset();
    IPC::S().put(CMD_FORK);
    IPC::S() << encode((int)getpid());
    IPC::S() << encode((int)getppid());
    IPC::Issue();
    delete IPC::Accept(); /* Wait for CMD_FORK to finish. */
  }
  return( pid );
}

int
MTCP::Socket(int family, int type, int protocol) {
  if (family != AF_RMP) {
    return( socket(family, type, protocol) );
  }
  int sockfd = IPC::ObtainConnection(CONNECTION_SOCKET);
  return sockfd;
}

int
MTCP::Connect(int sockfd, struct sockaddr* servaddr, int addrlen) {
  if (servaddr->sa_family != AF_RMP) {
    return( connect(sockfd, servaddr, addrlen) );
  }
  IPC::Reset();
  IPC::S().put(CMD_CONNECT);
  IPC::S() << encode(sockfd);
  IPC::S() << encode((int)getpid());
  IPC::S().write((char*)servaddr, addrlen);
  IPC::Issue();
  int rval;
  istrstream* msg = IPC::Accept();
  (*msg) >> decode(rval);
  delete msg;
  return rval;
}

int
MTCP::SetSockOpt(int sockfd, int level, int optname, char* optval, int optlen) {
  if (level != SOL_MTCP) {
    return( setsockopt(sockfd, level, optname, optval, optlen) );
  }
  IPC::Reset();
  IPC::S().put(CMD_SET_SOCK_OPT);
  IPC::S() << encode(sockfd);
  IPC::S() << encode((int)getpid());
  IPC::S() << encode(optname);
  IPC::S() << encode(optlen);
  IPC::S().write(optval, optlen);
  IPC::Issue();
  int rval;
  istrstream* msg = IPC::Accept();
  (*msg) >> decode(rval);
  delete msg;
  return rval;
}

int
MTCP::GetSockOpt(int sockfd, int level, int optname, char* optval, int* optlen) {
  if (level != SOL_MTCP) {
    return( getsockopt(sockfd, level, optname, optval, optlen) );
  }
  IPC::Reset();
  IPC::S().put(CMD_GET_SOCK_OPT);
  IPC::S() << encode(sockfd);
  IPC::S() << encode((int)getpid());
  IPC::S() << encode(optname);
  IPC::Issue();
  int rval;
  int length;
  istrstream* msg = IPC::Accept();
  (*msg) >> decode(rval);
  if (rval == 0) {
    (*msg) >> decode(length);
    *optlen = MIN2(*optlen, length);
    msg->read(optval,*optlen);
  }
  delete msg;
  return rval;
}

int
MTCP::GetPeerName(int sockfd, struct sockaddr* peer, int* addrlen) {
  return( GetSockName(sockfd, peer, addrlen) );
}

int
MTCP::GetSockName(int sockfd, struct sockaddr* peer, int* addrlen) {
  IPC::Reset();
  IPC::S().put(CMD_GET_SOCK_NAME);
  IPC::S() << encode(sockfd);
  IPC::S() << encode((int)getpid());
  IPC::Issue();
  int rval;
  istrstream* msg = IPC::Accept();
  (*msg) >> decode(rval);
  if (rval == 1) {
    delete msg;
    return( getsockname(sockfd, peer, addrlen) );
  }
  if (rval == 0) { // Success.
    int length;
    (*msg) >> decode(length);
    *addrlen = MIN2(*addrlen, length);
    msg->read((char*)peer, *addrlen);
  }
  delete msg;
  return rval;
}

int
MTCP::Close(int fd) {
  int rval = close(fd);
  if (rval == 0) {
    IPC::Reset();
    IPC::S().put(CMD_CLOSE);
    IPC::S() << encode(fd);
    IPC::S() << encode((int)getpid());
    IPC::Issue();
    istrstream* msg = IPC::Accept();
    (*msg) >> decode(rval);
    delete msg;
  }
  return rval;
}

int
MTCP::Bind(int sockfd, struct sockaddr* servaddr, int addrlen) {
  if (servaddr->sa_family != AF_RMP) {
    return( connect(sockfd, servaddr, addrlen) );
  }
  IPC::Reset();
  IPC::S().put(CMD_BIND);
  IPC::S() << encode(sockfd);
  IPC::S() << encode((int)getpid());
  IPC::S().write((char*)servaddr, addrlen);
  IPC::Issue();
  int rval;
  istrstream* msg = IPC::Accept();
  (*msg) >> decode(rval);
  delete msg;
  return rval;
}

int
MTCP::Accept(int sockfd, struct sockaddr* peer, int* addrlen) {
  IPC::Reset();
  IPC::S().put(CMD_ACCEPT);
  IPC::S() << encode(sockfd);
  IPC::S() << encode((int)getpid());
  IPC::Issue();
  int rval;
  istrstream* msg = IPC::Accept();
  (*msg) >> decode(rval);
  delete msg;
  if (rval == 0) { /* Valid MTCP_SOCKET FD. */
    delete IPCM::Accept(sockfd);
    if ( (peer != 0) && (addrlen != 0) ) {
      GetPeerName(sockfd, peer, addrlen);
    }
    return sockfd;
  }
  /* Not an MTCP_SOCKET FD. */
  return( accept(sockfd, peer, addrlen) );
}

void
MTCP::BeginMessage(int sockfd) {
  IPC::Reset();
  IPC::S().put(CMD_BEGIN_MESSAGE);
  IPC::S() << encode(sockfd);
  IPC::S() << encode((int)getpid());
  IPC::Issue();
  istrstream* msg = IPC::Accept();
  delete msg;
}

void
MTCP::EndMessage(int sockfd) {
  IPC::Reset();
  IPC::S().put(CMD_END_MESSAGE);
  IPC::S() << encode(sockfd);
  IPC::S() << encode((int)getpid());
  IPC::Issue();
  istrstream* msg = IPC::Accept();
  delete msg;
}

int
MTCP::Write(int sockfd, char* buff, unsigned int nbytes) {
  int rval;
  BeginMessage(sockfd);
  rval = write(sockfd, buff, nbytes);
  EndMessage(sockfd);
  return rval;
}
